from finalsa.dynamo.client.interface import SyncDynamoClient as Interface
from .implementation import DynamoClientImpl as Impl
from typing import List, Dict


class SyncDynamoClient(Interface):

    def __init__(self):
        self.__client__ = Impl()

    def write_transaction(self, transactions: List):
        self.__client__.write_transaction(transactions)

    def query(self, TableName: str, **kwargs):
        return self.__client__.query(TableName, **kwargs)

    def put(self, TableName: str, item: Dict):
        self.__client__.put(TableName, item)

    def delete(self, TableName: str, key: Dict):
        self.__client__.delete(TableName, key)

    def get(self, table_name: str, key: Dict) -> Dict:
        return self.__client__.get(table_name, key)

    def scan(self, TableName: str, **kwargs):
        return self.__client__.scan(TableName, **kwargs)



from finalsa.dynamo.client.interface import SyncDynamoClient, AsyncDynamoClient
from finalsa.dynamo.client.client import (
    SyncDynamoClient as SyncDynamoClientImpl, 
    AsyncDynamoClient as AsyncDynamoClientImpl
)
from finalsa.dynamo.client.tests import (SyncDynamoClientTestImpl, AsyncDynamoClientTestImpl)
from finalsa.dynamo.client.tests import seed



__version__ = "0.0.1"

__all__ = [
    "SyncDynamoClient",
    "SyncDynamoClientImpl",
    "SyncDynamoClientTestImpl",

    "AsyncDynamoClient",
    "AsyncDynamoClientImpl",
    "AsyncDynamoClientTestImpl",
    
    "seed"
]



from typing import List, Dict
from finalsa.dynamo.client.interface import SyncDynamoClient as Interface


def is_equal(item: Dict, key: Dict):
    for k, v in key.items():
        if item.get(k) != v:
            return False
    return True


class DynamoClientTestImpl(Interface):

    def __init__(self):
        self.tables = {}

    def get_table(self, TableName: str):
        if TableName not in self.tables:
            self.tables[TableName] = []
        return self.tables[TableName]

    def put_transaction(self, transaction: Dict):
        self.get_table(transaction['TableName'])
        self.tables[transaction['TableName']].append(transaction['Item'])

    def delete_transaction(self, transaction: Dict):
        self.get_table(transaction['TableName'])
        self.tables[transaction['TableName']] = list(filter(
            lambda x: x != transaction['Item'], self.tables[transaction['TableName']]))

    def update_transaction(self, transaction: Dict):
        self.get_table(transaction['TableName'])
        for item in self.tables[transaction['TableName']]:
            if is_equal(item, transaction['Key']):
                update_expression = transaction['UpdateExpression']
                update_expression = update_expression.replace('SET', '')
                update_expression = update_expression.replace(' ', '')
                update_expression = update_expression.split(',')
                update_expression_dict = {}
                for expression in update_expression:
                    key, value = expression.split('=')
                    update_expression_dict[key] = value
                for key, value in update_expression_dict.items():
                    attr = transaction["ExpressionAttributeNames"][key]
                    val = transaction["ExpressionAttributeValues"][value]
                    item[attr] = val
                return
        raise Exception('Item not found')

    def write_transaction(self, transactions: List):
        accepted_transaction_types = ['Put', 'Update', 'Delete']
        for transaction in transactions:
            for key in transaction:
                item = transaction[key]
                if key not in accepted_transaction_types:
                    raise Exception('Invalid transaction type')
                if key == 'Put':
                    self.put_transaction(item)
                elif key == 'Update':
                    self.update_transaction(item)
                elif key == 'Delete':
                    self.delete_transaction(item)

    def query(self, TableName: str, **kwargs):
        if TableName not in self.tables:
            return {'Items': [], 'Count': 0}
        expression_attribute_values = kwargs.get(
            'ExpressionAttributeValues', {})
        key_condition_expression = kwargs.get('KeyConditionExpression', '')
        key_condition_expression = key_condition_expression.replace(' ', '')
        key_condition_expression = key_condition_expression.split('AND')
        key_condition_expression_dict = {}
        for expression in key_condition_expression:
            key, value = expression.split('=')
            key_condition_expression_dict[key] = expression_attribute_values[
                value]
        items = []
        for item in self.tables[TableName]:
            if is_equal(item, key_condition_expression_dict):
                items.append(item)
        return {'Items': items, 'Count': len(items)}

    def get(self, TableName: str, key: Dict):
        self.get_table(TableName)
        for i in self.tables[TableName]:
            if is_equal(i, key):
                return i
        return None

    def put(self, TableName: str, item: Dict):
        self.get_table(TableName)
        self.tables[TableName].append(item)

    def delete(self, TableName: str, key: Dict):
        self.get_table(TableName)
        self.tables[TableName] = list(filter(
            lambda x: not is_equal(x, key), self.tables[TableName]))

    def scan(self, TableName: str, **kwargs):
        if TableName not in self.tables:
            return {'Items': []}

        def begins_with(x: str, y):
            return x.startswith(y)
        actions = {
            "begins_with": begins_with
        }
        if not kwargs or len(kwargs) == 0:
            return {'Items': self.tables[TableName]}
        filters = []
        if 'FilterExpression' in kwargs:
            expresions = kwargs['FilterExpression']
            expresions = expresions.replace(' ', '')
            expresions = expresions.split('AND')
            values = kwargs['ExpressionAttributeValues']

            for expresion in expresions:
                if 'begins_with' in expresion:
                    expresion = expresion.replace('begins_with', '')
                    expresion = expresion.replace('(', '')
                    expresion = expresion.replace(')', '')
                    key, value = expresion.split(',')
                    filters.append(('begins_with', {
                        "key": key,
                        "value": values[value]
                    }))

        items = []
        for item in self.tables[TableName]:
            for filter in filters:
                if filter[0] in actions:
                    action_eval = actions[filter[0]](item[filter[1]['key']], filter[1]['value'])
                    if action_eval:
                        items.append(item)
        return {'Items': items}

    def update(self, TableName: str, key: Dict, item: Dict):
        self.get_table(TableName)
        for i in self.tables[TableName]:
            if is_equal(i, key):
                for k, v in item.items():
                    i[k] = v
                return i
        return {'Item': None}



from .client import SyncDynamoClientTestImpl
from .async_client import AsyncDynamoClientTestImpl
from typing import List, Dict, Union


def seed(
    table: str,
    items: List[Dict],
    client: Union[AsyncDynamoClientTestImpl,
                      SyncDynamoClientTestImpl] = SyncDynamoClientTestImpl()
):
    for item in items:
        client.__impl__.put(TableName=table, item=item)



import boto3
from finalsa.dynamo.client.interface import SyncDynamoClient
from typing import List, Dict


class DynamoClientImpl(SyncDynamoClient):

    def __init__(self):
        self.client = boto3.client("dynamodb")

    def __write_transaction__(self, transactions: List):
        self.client.transact_write_items(TransactItems=transactions)

    def __query__(self, TableName: str, **kwargs):
        return self.client.query(TableName=TableName, **kwargs)

    def __put__(self, TableName: str, item: Dict):
        self.client.put_item(TableName=TableName, Item=item)

    def __get__(self, TableName: str, key: Dict):
        return self.client.get_item(TableName=TableName, Key=key)

    def __delete__(self, TableName: str, key: Dict):
        self.client.delete_item(TableName=TableName, Key=key)

    def __scan__(self, TableName: str, **kwargs):
        return self.client.scan(TableName=TableName, **kwargs)



from finalsa.dynamo.client import (SyncDynamoClientTestImpl, __version__)
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../../')))


def test_version():
    assert __version__ is not None


def test_dynamo_client_test_impl():
    client = SyncDynamoClientTestImpl()
    assert client is not None
    assert client.put is not None
    assert client.query is not None
    assert client.delete is not None
    assert client.scan is not None
    assert client.write_transaction is not None
    assert client.get is not None


def test_dynamo_client_put():
    client = SyncDynamoClientTestImpl()
    client.put("test", {"id": "1", "name": "test"})


def test_dynamo_client_query():
    client = SyncDynamoClientTestImpl()
    client.put("test", {"id": "1", "name": "test"})
    client.put("test", {"id": "2", "name": "test"})
    client.put("test", {"id": "3", "name": "test"})
    assert client.get("test", {"id": "1"}) is not None
    assert client.get("test", {"id": "2"}) is not None
    assert client.get("test", {"id": "3"}) is not None


def test_dynamo_client_delete():
    client = SyncDynamoClientTestImpl()
    client.put("test", {"id": "1", "name": "test"})
    client.put("test", {"id": "2", "name": "test"})
    client.put("test", {"id": "3", "name": "test"})
    assert client.get("test", {"id": "1"}) is not None
    assert client.get("test", {"id": "2"}) is not None
    assert client.get("test", {"id": "3"}) is not None
    client.delete("test", {"id": "1"})
    assert client.get("test", {"id": "1"}) is None


def test_dynamo_client_scan():
    client = SyncDynamoClientTestImpl()
    client.put("test", {"id": "a1", "name": "test"})
    client.put("test", {"id": "a2", "name": "test"})
    client.put("test", {"id": "b3", "name": "test"})
    assert client.scan("test", FilterExpression="begins_with(id,:id)", ExpressionAttributeValues={
        ":id": "a"
    }) is not None
    assert client.scan("test").get("Items") is not None
    assert len(client.scan("test",  FilterExpression="begins_with(id,:id)", ExpressionAttributeValues={
        ":id": "a"
    }).get("Items")) == 2
    assert len(client.scan("test").get("Items")) == 3


def test_dynamo_client_write_transaction():
    client = SyncDynamoClientTestImpl()
    client.write_transaction(
        [{"Put": {"TableName": "test", "Item": {"id": "1", "name": "test"}}}])
    assert client.get("test", {"id": "1"}) is not None



from abc import ABC, abstractmethod
from typing import List, Dict


class SyncDynamoClient(ABC):

    @abstractmethod
    def write_transaction(self, transactions: List):
        pass

    @abstractmethod
    def query(self, table_name: str, **kwargs):
        pass

    @abstractmethod
    def put(self, table_name: str, item: Dict):
        pass

    @abstractmethod
    def delete(self, table_name: str, key: Dict):
        pass

    @abstractmethod
    def scan(self, table_name: str, **kwargs):
        pass

    @abstractmethod
    def get(self, table_name: str, key: Dict) -> Dict:
        pass

    @abstractmethod
    def update(self, table_name: str, key: Dict, update_expression: str, expression_attribute_values: Dict):
        pass



from typing import List, Dict
from finalsa.dynamo.client.interface import AsyncDynamoClient as Interface
from .implementation import DynamoClientTestImpl as Impl


class AsyncDynamoClientTestImpl(Interface):

    def __init__(self):
        self.__impl__ = Impl()

    async def write_transaction(self, transactions: List):
        self.__impl__.write_transaction(transactions)

    async def query(self, TableName: str, **kwargs):
        return self.__impl__.query(TableName, **kwargs)

    async def get(self, TableName: str, item: Dict):
        return self.__impl__.get(TableName, item)

    async def put(self, TableName: str, item: Dict):
        self.__impl__.put(TableName, item)

    async def delete(self, TableName: str, key: Dict):
        self.__impl__.delete(TableName, key)

    async def scan(self, TableName: str, **kwargs):
        return self.__impl__.scan(TableName, **kwargs)

    async def update(self, TableName: str, key: Dict, item: Dict):
        return self.__impl__.update(TableName, key, item)
